import sqlite3
from flask import current_app, render_template

from chill.app import db
from database import fetch_query_string, rowify

def _short_circuit(value=None):
    """
    Add the `value` to the `collection` by modifying the collection to be
    either a dict or list depending on what is already in the collection and
    value.
    Returns the collection with the value added to it.

    Clean up by removing single item array and single key dict.
    ['abc'] -> 'abc'
    [['abc']] -> 'abc'
    [{'abc':123},{'def':456}] -> {'abc':123,'def':456}
    [{'abc':123},{'abc':456}] -> [{'abc':123,'abc':456}] # skip for same set keys
    [[{'abc':123},{'abc':456}]] -> [{'abc':123,'abc':456}]
    """
    if not isinstance(value, list):
        return value
    if len(value) == 0:
        return value
    if len(value) == 1:
        if not isinstance(value[0], list):
            return value[0]
        else:
            if len(value[0]) == 1:
                return value[0][0]
            else:
                return value[0]
    else:
        value = filter(None, value)
        # Only checking first item and assumin all others are same type
        if isinstance(value[0], dict):
            if set(value[0].keys()) == set(value[1].keys()):
                return value
            elif max([len(x.keys()) for x in value]) == 1:
                newvalue = {}
                for v in value:
                    key = v.keys()[0]
                    newvalue[key] = v[key]
                return newvalue
            else:
                return value
        else:
            return value



def _query(_node_id, value=None, **kw):
    "Look up value by using Query table"
    c = db.cursor()
    try:
        result = c.execute(fetch_query_string('select_query_from_node.sql'), kw).fetchall()
    except sqlite3.DatabaseError as err:
        current_app.logger.error("DatabaseError: %s, %s", err, kw)
        return value
    (query_result, query_col_names) = rowify(result, c.description)
    #current_app.logger.debug("queries kw: %s", kw)
    #current_app.logger.debug("queries value: %s", value)
    #current_app.logger.debug("queries: %s", query_result)
    if query_result:
        values = []
        for query_name in [x.get('name', None) for x in query_result]:
            if query_name:
                try:
                    result = c.execute(fetch_query_string(query_name), kw).fetchall()
                    values.append( rowify(result, c.description) )
                except sqlite3.DatabaseError as err:
                    current_app.logger.error("DatabaseError (%s) %s: %s", query_name, kw, err)
        value = values
    #current_app.logger.debug("value: %s", value)
    return value

def _link(node_id):
    "Add the value for a linked node"
    c = db.cursor()
    linked_value = c.execute(fetch_query_string('select_link_node_from_node.sql'), {'node_id': node_id}).fetchall()
    if linked_value:
        if len(linked_value) > 1:
            list = []
            for v in linked_value:
                list.append({v[1]: render_node(v[2], None, v[1])})
            linked_value = list
        else:
            linked_value = render_node(linked_value[0][0]) #TODO
    return linked_value

def _template(node_id, value=None):
    "Check if a template is assigned to it and render that with the value"
    c = db.cursor()
    select_template_from_node = fetch_query_string('select_template_from_node.sql')
    try:
        c.execute(select_template_from_node, {'node_id':node_id})
        template_result = c.fetchone()
        if template_result and template_result[1]:
            template = template_result[1]

            if isinstance(value, dict):
                return render_template(template, **value)
            else:
                return render_template(template, value=value)
    except sqlite3.DatabaseError as err:
        current_app.logger.error("DatabaseError: %s", err)

    # No template assigned to this node so just return the value
    return value

def render_node(_node_id, value=None, noderequest={}, **kw):
    "Recursively render a node's value"
    if value == None:
        kw.update( noderequest )
        results = _query(_node_id, **kw)
        if results and results[0]:
            values = []
            for (result, cols) in results:
                if set(cols) == set(['node_id', 'name', 'value']):
                    for subresult in result:
                        #if subresult.get('name') == kw.get('name'):
                            # This is a link node
                        #current_app.logger.debug("sub: %s", subresult)
                        name = subresult.get('name')
                        if noderequest.get('_no_template'):
                            # For debugging or just simply viewing with the
                            # operate script we append the node_id to the name
                            # of each. This doesn't work with templates.
                            name = "{0} - {1}".format(name, subresult.get('node_id'))
                        values.append( {name: render_node( subresult.get('node_id'), noderequest=noderequest, **subresult )} )
                #elif 'node_id' and 'name' in cols:
                #    for subresult in result:
                #        current_app.logger.debug("sub2: %s", subresult)
                #        values.append( {subresult.get('name'): render_node( subresult.get('node_id'), **subresult )} )
                else:
                    values.append( result )

            value = values

    value = _short_circuit(value)
    if not noderequest.get('_no_template'):
        value = _template(_node_id, value)

    return value



"""Operate script to edit a SQL Database that is used by chill.

Some various operations are selected via a select menu.  Entering the number
and pressing return will select that item and will show other prompts to
interactively edit the content in the database. Entering nothing and simply
hitting return will exit that menu.

All of these operations can be manually done via SQL, but this is more of
a simple approach using functions found within chill.  If needing to add lots
of content it would be wise to write a script to handle these things.

"""
import os
from glob import glob
import re

from yaml import safe_dump
import sqlite3
from flask import current_app
from pyselect import select
from chill.app import db
from api import render_node
from chill.database import (
        init_db,
        init_picture_tables,
        insert_node,
        insert_node_node,
        insert_route,
        insert_query,
        add_template_for_node,
        fetch_query_string,
        add_picture_for_node,
        link_picturename_for_node,
        rowify,
        )

INVALID_NODE = -1

def node_input():
    """
    Get a valid node id from the user.

    Return -1 if invalid
    """
    try:
        node = int(raw_input("Node id: "))
    except ValueError:
        node = INVALID_NODE
        print 'invalid node id: %s' % node
    return node

def existing_node_input():
    """
    Get an existing node id by name or id.

    Return -1 if invalid
    """
    input_from_user = raw_input("Existing node name or id: ")
    node_id = INVALID_NODE
    c = db.cursor()

    if not input_from_user:
        return node_id

    # int or str?
    try:
        parsed_input = int(input_from_user)
    except ValueError:
        parsed_input = input_from_user

    if isinstance(parsed_input, int):
        c.execute(fetch_query_string('select_node_from_id.sql'),
                {'node_id':parsed_input})
        result = c.fetchall()
        (result, col_names) = rowify(result, c.description)
        current_app.logger.debug("result: %s", result)
        if result:
            node_id = int(result[0].get('node_id'))
    else:
        c.execute(fetch_query_string('select_node_from_name.sql'),
                {'node_name':parsed_input})
        result = c.fetchall()
        (result, col_names) = rowify(result, c.description)
        if result:
            if len(result) == 1:
                print 'Node id: {node_id}\nNode name: {name}'.format(**result[0])
                print '-------------'
                node_id = result[0].get('node_id')
            else:
                print 'Multiple nodes found with the name: {0}'.format(parsed_input)
                for item in result:
                    print '{node_id}: {name} = {value}'.format(**item)
                node_selection = raw_input('Enter a node id from this list or enter "?" to render all or "?<node>" for a specific one.')
                if node_selection:
                    node_selection_match = re.match(r"\?(\d)*", node_selection)
                    if node_selection_match:
                        if node_selection_match.groups()[0]:
                            value = render_node(int(node_selection_match.groups()[0]), noderequest={'_no_template':True}, **result[0])
                            print safe_dump(value, default_flow_style=False)
                        else:
                            for item in result:
                                value = render_node(item.get('node_id'), noderequest={'_no_template':True}, **item)
                                print 'Node id: {0}'.format(item.get('node_id'))
                                print safe_dump(value, default_flow_style=False)
                                print '---'
                        node_id = node_input()
                    else:
                        try:
                            node_id = int(node_selection)
                        except ValueError:
                            node_id = INVALID_NODE
                            print 'invalid node id: %s' % node

    return node_id

def choose_query_file():
    print "Choose from the available query files:"
    choices = set(
            map(os.path.basename,
                glob(os.path.join(os.path.dirname(__file__), 'queries', '*'))
                )
            )
    folder = current_app.config.get('THEME_SQL_FOLDER')
    choices.update(
            set(map(os.path.basename,
                glob(os.path.join(folder, '*'))
                )
            ))
    choices = list(choices)
    choices.sort()
    return select(choices)

def mode_collection():
    "Create a collection of items with common attributes"

    print globals()['mode_collection'].__doc__
    collection_name = raw_input("Collection name: ")
    item_attr_list = []
    if collection_name:
        collection_node_id = insert_node(name=collection_name, value=None)
        insert_query(name='select_link_node_from_node.sql', node_id=collection_node_id)
        item_attr = True
        while item_attr:
            item_attr = raw_input("Add a collection item attribute name: ")
            if item_attr:
                item_attr_list.append(item_attr)

    # if no collection name then exit
    selection = collection_name

    while selection:
        selection = select([
            'Add item',
            ])
        if selection == 'Add item':
            # create item
            item_node_id = insert_node(name='{0}-item'.format(collection_name), value=None)
            insert_query(name='select_link_node_from_node.sql', node_id=item_node_id)
            insert_node_node(node_id=collection_node_id, target_node_id=item_node_id)
            for item_attr_name in item_attr_list:
                value = raw_input("Enter item attribute value for '{0}': ".format(item_attr_name))
                # set value to none if it's an empty string
                value = value if len(value) else None
                item_attr_node_id = insert_node(name=item_attr_name, value=value)
                insert_node_node(node_id=item_node_id, target_node_id=item_attr_node_id)

    if collection_node_id:
        print "Added collection name '{0}' with node id: {1}".format(collection_name, collection_node_id)

    db.commit()


def mode_database_functions():
    "Select a function to perform from chill.database"

    print globals()['mode_database_functions'].__doc__
    selection = True
    database_functions = [
            'init_db',
            'init_picture_tables',
            'insert_node',
            'insert_node_node',
            'insert_route',
            'insert_query',
            'add_template_for_node',
            'add_picture_for_node',
            'link_picturename_for_node',
            ]
    while selection:
        choices = database_functions + [
            'help',
            ]
        selection = select(choices)

        if selection:
            print globals().get(selection).__doc__
        if selection == 'init_db':
            confirm = raw_input("Initialize new database y/n? [n] ")
            if confirm == 'y':
                init_db()
        if selection == 'init_picture_tables':
            confirm = raw_input("Initialize new tables for pictures y/n? [n] ")
            if confirm == 'y':
                init_picture_tables()
        elif selection == 'insert_node':
            name = raw_input("Node name: ")
            value = raw_input("Node value: ")
            node = insert_node(name=name, value=value or None)
            print "name: %s \nid: %s" % (name, node)

        elif selection == 'insert_query':
            sqlfile = choose_query_file()
            if sqlfile:
                node = existing_node_input()
                if node >= 0:
                    insert_query(name=sqlfile, node_id=node)
                    print "adding %s to node id: %s" % (sqlfile, node)

        elif selection == 'insert_node_node':
            print "Add parent node id"
            node = existing_node_input()
            print "Add target node id"
            target_node = existing_node_input()
            if node >= 0 and target_node >= 0:
                insert_node_node(node_id=node, target_node_id=target_node)

        elif selection == 'insert_route':
            path = raw_input('path: ')
            weight = raw_input('weight: ')
            method = raw_input('method: ') or 'GET'
            node = existing_node_input()
            if node >= 0:
                insert_route(path=path, node_id=node, weight=weight, method=method)
        elif selection == 'add_template_for_node':
            folder = current_app.config.get('THEME_TEMPLATE_FOLDER')
            choices = map(os.path.basename,
                        glob(os.path.join(folder, '*'))
                        )
            choices.sort()
            templatefile = select(choices)
            if templatefile:
                node = existing_node_input()
                if node >= 0:
                    add_template_for_node(name=templatefile, node_id=node)
                    print "adding %s to node id: %s" % (templatefile, node)
        elif selection == 'add_picture_for_node':
            if current_app.config.get('MEDIA_FOLDER'):
                node = existing_node_input()

                if node >= 0:
                    filepath = raw_input("Enter the filepath in the media folder. Enter nothing to bring up a list.")
                    if not filepath:
                        filelist = glob(os.path.join(current_app.config.get('MEDIA_FOLDER'), '*'))
                        if len(filelist) > 0:
                            toplevel_files = map(os.path.basename, filelist)
                            toplevel_files.sort()
                            filepath = select(toplevel_files)
                        else:
                            print "no files found in media folder."
                    if filepath:
                        add_picture_for_node(node_id=node, filepath=filepath)

        elif selection == 'link_picturename_for_node':
            if current_app.config.get('MEDIA_FOLDER'):
                node = existing_node_input()

                if node >= 0:
                    filepath = raw_input("Enter the filepath/picturename in the media folder. Enter nothing to bring up a list.")
                    if not filepath:
                        filelist = glob(os.path.join(current_app.config.get('MEDIA_FOLDER'), '*'))
                        if len(filelist) > 0:
                            toplevel_files = map(os.path.basename, filelist)
                            toplevel_files.sort()
                            filepath = select(toplevel_files)
                        else:
                            print "no files found in media folder."
                    if filepath:
                        link_picturename_for_node(node_id=node, picturename=filepath)

        elif selection == 'help':
            print "------"
            for f in database_functions:
                print "\n** %s **" % f
                print globals().get(f).__doc__
            print "------"
        else:
            pass

        db.commit()

def operate_menu():
    "Select between these operations on the database"

    selection = True
    while selection:

        print globals()['operate_menu'].__doc__
        selection = select([
            'chill.database functions',
            'execute sql file',
            'render_node',
            'Create collection',
            'Add document for node',
            'help',
            ])
        if selection == 'chill.database functions':
            mode_database_functions()
        elif selection == 'execute sql file':
            print "View the sql file and show a fill in the blanks interface with raw_input"
            sqlfile = choose_query_file()
            if not sqlfile:
                # return to the menu choices if not file picked
                selection = True
            else:
                sql_named_placeholders_re = re.compile(r":(\w+)")
                sql = fetch_query_string(sqlfile)
                placeholders = set(sql_named_placeholders_re.findall(sql))
                print sql
                data = {}
                for placeholder in placeholders:
                    value = raw_input(placeholder + ': ')
                    data[placeholder] = value

                c = db.cursor()
                try:
                    c.execute(sql, data)
                except sqlite3.DatabaseError as err:
                    current_app.logger.error("DatabaseError: %s", err)

                result = c.fetchall()
                if result:
                    (result, col_names) = rowify(result, c.description)
                    kw = result[0]

                    if 'node_id' in kw:
                        value = render_node(kw['node_id'], **kw)
                        print safe_dump(value, default_flow_style=False)
                    else:
                        print safe_dump(result, default_flow_style=False)

        elif selection == 'render_node':
            print globals()['render_node'].__doc__
            node_id = existing_node_input()

            c = db.cursor()
            try:
                c.execute(fetch_query_string('select_node_from_id.sql'), {'node_id':node_id})
            except sqlite3.DatabaseError as err:
                current_app.logger.error("DatabaseError: %s", err)

            result = c.fetchall()
            if result:
                (result, col_names) = rowify(result, c.description)
                kw = result[0]

                value = render_node(node_id, noderequest={'_no_template':True}, **kw)
                print safe_dump(value, default_flow_style=False)

        elif selection == 'Create collection':
            mode_collection()
        elif selection == 'Add document for node':
            folder = current_app.config.get('DOCUMENT_FOLDER')
            if not folder:
                print "No DOCUMENT_FOLDER configured for the application."
            else:
                choices = map(os.path.basename,
                            glob(os.path.join(folder, '*'))
                            )
                choices.sort()
                if len(choices) == 0:
                    print "No files found in DOCUMENT_FOLDER."
                else:
                    filename = select(choices)
                    if filename:
                        defaultname = os.path.splitext(filename)[0]
                        nodename = raw_input("Enter name for node [{0}]: ".format(defaultname)) or defaultname
                        node = insert_node(name=nodename, value=filename)
                        print "Added document '%s' to node '%s' with id: %s" % (filename, nodename, node)
        elif selection == 'help':
            print "------"
            print __doc__
            print "------"
        else:
            print 'Done'

        db.commit()





import os
from flask import current_app, g
from chill.app import db
from cache import cache
from PIL import Image

CHILL_CREATE_TABLE_FILES = (
        'create_chill.sql',
        'create_node_node.sql',
        'create_node.sql',
        'create_route.sql',
        'create_query.sql',
        'create_template.sql'
        )

CHILL_CREATE_PICTURE_TABLE_FILES = (
    'create_picture.sql',
    'create_image.sql',
    'create_srcset.sql',
    'create_staticfile.sql',
    'create_node_picture.sql'
    )


def init_db():
    """Initialize a new database with the default tables for chill.
    Creates the following tables:
    Chill
    Node
    Node_Node
    Route
    Query
    Template
    """
    with current_app.app_context():
        #db = get_db()
        c = db.cursor()

        for filename in CHILL_CREATE_TABLE_FILES:
            c.execute(fetch_query_string(filename))

        db.commit()

def rowify(l, description):
    d = []
    col_names = []
    if l != None and description != None:
        col_names = [x[0] for x in description]
        for row in l:
            d.append(dict(zip(col_names, row)))
    return (d, col_names)

def _fetch_sql_string(file_name):
    with current_app.open_resource(os.path.join('queries', file_name), mode='r') as f:
        return f.read()

def fetch_query_string(file_name):
    content = current_app.queries.get(file_name, None)
    if content != None:
        return content
    current_app.logger.info( "queries file: '%s' not available. Checking file system..." % file_name )

    #folder = current_app.config.get('THEME_SQL_FOLDER', '')
    #file_path = os.path.join(os.path.abspath('.'), folder, file_name)
    folder = current_app.config.get('THEME_SQL_FOLDER')
    file_path = os.path.join(folder, file_name)
    if os.path.isfile(file_path):
        with open(file_path, 'r') as f:
            return f.read()
    else:
        # fallback on one that's in app resources
        return _fetch_sql_string(file_name)

def insert_node(**kw):
    "Insert a node with a name and optional value. Return the node id."
    with current_app.app_context():
        c = db.cursor()
        c.execute(fetch_query_string('insert_node.sql'), kw)
        node_id = c.lastrowid
        db.commit()
        return node_id

def insert_node_node(**kw):
    """
    Link a node to another node. node_id -> target_node_id.  Where `node_id` is
    the parent and `target_node_id` is the child.
    """
    with current_app.app_context():
        insert_query(name='select_link_node_from_node.sql', node_id=kw.get('node_id'))
        c = db.cursor()
        c.execute(fetch_query_string('insert_node_node.sql'), kw)
        db.commit()


def insert_route(**kw):
    """
    `path` - '/', '/some/other/path/', '/test/<int:index>/'
    `node_id`
    `weight` - How this path is selected before other similar paths
    `method` - 'GET' is default.
    """
    binding = {
            'path': None,
            'node_id': None,
            'weight': None,
            'method': "GET"
            }
    binding.update(kw)
    with current_app.app_context():
        c = db.cursor()
        c.execute(fetch_query_string('insert_route.sql'), binding)
        db.commit()

def add_template_for_node(name, node_id):
    "Set the template to use to display the node"
    with current_app.app_context():
        c = db.cursor()
        c.execute(fetch_query_string('insert_template.sql'),
                {'name':name, 'node_id':node_id})
        c.execute(fetch_query_string('select_template.sql'),
                {'name':name, 'node_id':node_id})
        result = c.fetchone()
        if result:
            template_id = result[0]
            c.execute(fetch_query_string('update_template_node.sql'),
                    {'template':template_id, 'node_id':node_id})
        db.commit()


def insert_query(**kw):
    """
    Insert a query name for a node_id.
    `name`
    `node_id`

    Adds the name to the Query table if not already there. Sets the query field
    in Node table.
    """
    with current_app.app_context():
        c = db.cursor()
        result = c.execute(fetch_query_string('select_query_where_name.sql'), kw).fetchall()
        (result, col_names) = rowify(result, c.description)
        if result:
            kw['query_id'] = result[0].get('id')
        else:
            c.execute(fetch_query_string('insert_query.sql'), kw)
            kw['query_id'] = c.lastrowid
        c.execute(fetch_query_string('insert_query_node.sql'), kw)
        db.commit()

def init_picture_tables():
    """Create optional picture and staticfile database tables:
    Picture
    Image
    Srcset
    StaticFile
    Node_Picture
    """
    with current_app.app_context():
        c = db.cursor()

        for filename in CHILL_CREATE_PICTURE_TABLE_FILES:
            c.execute(fetch_query_string(filename))

        db.commit()


def add_picture_for_node(node_id, filepath, **kw):
    """
    Add a picture for a node id. This adds to the Image, StaticFile, Picture, ... tables.
    The `filepath` must be an image file within the media folder.
    width and height are deduced from the image.
    Other attributes that should be associated with the picture can be passed in:
    title
    description
    author
    (and others, some have not been implemented)
    """

    with current_app.app_context():
        c = db.cursor()


        # media folder needs to be set
        media_folder = current_app.config.get('MEDIA_FOLDER')
        if not media_folder:
            current_app.logger.warn('No MEDIA_FOLDER set in config.')
            return False

        # filepath needs to exist
        media_filepath = os.path.join(media_folder, filepath)
        if not os.path.exists(media_filepath):
            current_app.logger.warn('filepath not exists: {0}'.format(media_filepath))
            return False

        # file needs to be an image
        try:
            img = Image.open(media_filepath)
        except IOError as err:
            current_app.logger.warn(err)
            return False


        (width, height) = img.size


        c.execute(fetch_query_string("insert_staticfile.sql"), {
            'path':filepath
            })
        staticfile = c.lastrowid

        c.execute(fetch_query_string("insert_image.sql"),{
            'width': width,
            'height': height,
            'staticfile': staticfile
            })
        image = c.lastrowid

        c.execute(fetch_query_string("insert_picture.sql"),{
            'picturename': filepath,
            'title': kw.get('title', None),
            'description': '',
            'author': None,
            'created': '',
            'image': image
            })
        picture = c.lastrowid

        c.execute(fetch_query_string("insert_node_picture.sql"),{
            'node_id': node_id,
            'picture': picture
            })

        db.commit()

        insert_query(name='select_picture_for_node.sql', node_id=node_id)

        db.commit()

def link_picturename_for_node(node_id, picturename, **kw):
    """
    Link a picture for a node id.  Use this if the picture has already been added to the database.
    """
    with current_app.app_context():
        c = db.cursor()

        result = c.execute(fetch_query_string("select_picture_by_name.sql"), {
            'picturename':picturename
            })
        (result, col_names) = rowify(result, c.description)
        if result:
            picture = result[0].get('id')
        else:
            current_app.logger.warn('picture by name:"{0}" is not in database.'.format(filepath))
            return False

        c.execute(fetch_query_string("insert_node_picture.sql"),{
            'node_id': node_id,
            'picture': picture
            })

        db.commit()

        insert_query(name='select_picture_for_node.sql', node_id=node_id)

        db.commit()
